import re
import string

def space_punctuations(input:str)->str:
    """
        This Function returns a string after removing unnecessary spaces before and after punctuations in an
        Urdu string.
    
        Arguments:
                    input (str): String that needs to be processed
        Returns:
                    input (string): String after unnecessary spaces have been removed before and after puntuations.
        Examples:
            >>> text1 = ' میں ۱٦ سال کا ہوں ؟ '
            >>> text1 = normalize_characters(text1)
            >>> print(text1)
            میں ۱٦ سال کا ہوں؟
            >>> text2 = 'میرا نام دانیال ہے ، میں مقصد لیب میں کام کرتا ہوں ۔'
            >>> text2 = normalize_characters(text2)
            >>> print(text2)
            میرا نام دانیال ہے، میں مقصد لیب میں کام کرتا ہوں۔
    """
    
    all_urdu_characters = {'\u0600':'\u0600','\u0601':'\u0601','\u0602':'\u0602','\u0603':'\u0603','،':'\u060c','؍':'\u060d','؎':'\u060e','؏':'\u060f','ؐ':'\u0610','ؑ':'\u0611','ؒ':'\u0612','ؓ':'\u0613','ؔ':'\u0614','ؕ':'\u0615','؛':'\u061b','؟':'\u061f','ء':'\u0621','ً':'\u064b','ٌ':'\u064c','ٍ':'\u064d','َ':'\u064e','ُ':'\u064f','ِ':'\u0650','ّ':'\u0651','ْ':'\u0652','ٓ':'\u0653','ٔ':'\u0654','ٖ':'\u0656','ٗ':'\u0657','٘':'\u0658','٪':'\u066a','٫':'\u066b','٬':'\u066c','ٰ':'\u0670','۔':'\u06d4','آ':'\u0622','أ':'\u0623','ا':'\u0627','ب':'\u0628','پ':'\u067e','ت':'\u062a','ٹ':'\u0679','ث':'\u062b','ج':'\u062c','چ':'\u0686','ح':'\u062d','خ':'\u062e','د':'\u062f','ڈ':'\u0688','ذ':'\u0630','ر':'\u0631','ڑ':'\u0691','ز':'\u0632','ژ':'\u0698','س':'\u0633','ش':'\u0634','ص':'\u0635','ض':'\u0636','ط':'\u0637','ظ':'\u0638','ع':'\u0639','غ':'\u063a','ف':'\u0641','ق':'\u0642','ک':'\u06a9','گ':'\u06af','ل':'\u0644','م':'\u0645','ن':'\u0646','ں':'\u06ba','و':'\u0648','ؤ':'\u0624','ھ':'\u06be','ہ':'\u06c1','ۂ':'\u06c2','ۃ':'\u06c3','ی':'\u06cc','ئ':'\u0626','ے':'\u06d2','ۓ':'\u06d3','۰':'\u06f0','۱':'\u06f1','۲':'\u06f2','۳':'\u06f3','۴':'\u06f4','۵':'\u06f5','۶':'\u06f6','۷':'\u06f7','۸':'\u06f8','۹':'\u06f9',}
    punctuations = "؛ ، ٫  ؟ ۔ ٪"+''.join(string.punctuation)
    space_after = re.compile(
        r"(?<=[" + "".join(punctuations) + "])(?=[^" + "".join(punctuations) + "0-9 \n])",
        flags=re.U | re.M | re.I)
    space_before = re.compile(r'\s+([' + "".join(punctuations) + '])',
                                                  flags=re.U | re.M | re.I
                                                  )
    return space_before.sub(r'\1', space_after.sub(' ',input))

def replace_digits(input:str,ToEng = True)->str:
    """
            This Function returns a string after replacing Urdu digits in a string with English 
            or English digits with Urdu digits.
           
            Arguments:
                    input (str): String that needs to be processed
                    ToEng (bool): Used to select the mode deciding whether to replace Urdu digits with English
                        or English digits with urdu, default set to True.
                        When True: replaces Urdu digits to English digits
                        When False: replaces English digits to Urdu digits
            Returns:
                    input (string): String from which digits have been replaced
            Examples:
                >>> text1 = 'یہ ۲۰۲۲ ہے۔'
                >>> text1 = replace_digits(text1)
                >>> print(text1)
                یہ 2022 ہے
                >>> text2 = "this is Python 3.10"
                >>> text2 = replace_digits(text2,False)
                >>> print(text2)
                this is Python ۳.۱۰      
    """
    english_to_urdu_digits_dict = {'0': '۰', '1': '۱', '2': '۲', '3': '۳', '4': '۴', '5': '۵', '6': '۶', '7': '۷', '8': '۸', '9': '۹'}
    for key,val in english_to_urdu_digits_dict.items():
        if ToEng == True:
            input = input.replace(val, key)
        else:
            input = input.replace(key, val)
    return input

def normalize_characters(input:str)->str:
    """
        This Function returns a string after normalizing by bringing all the characters in range of
        urdu Unicode characters.
    
        Arguments:
                input (str): String that needs to be processed
        Returns:
                input (string): String after bringing all characters into Urdu Unicode.
        Examples:
            >>> text1 = 'ﺂن اپنا رویہ درست کریں'
            >>> text1 = normalize_characters(text1)
            >>> print(text1)
            آن اپنا رویہ درست کریں
            >>> text2 = 'میں ۱٦ سال کا ہوں'
            >>> text2 = normalize_characters(text2)
            >>> print(text2)
            میں ۱۶ سال کا ہوں
    """
    urdu_character_mapping = {'آ':['ﺁ','ﺂ'],'أ':['ﺃ'],'ا':['ﺍ','ﺎ',],'ب':['ﺏ','ﺐ','ﺑ','ﺒ'],'پ':['ﭖ','ﭘ','ﭙ',],'ت':['ﺕ','ﺖ','ﺗ','ﺘ'],'ٹ':['ﭦ','ﭧ','ﭨ','ﭩ'],'ث':['ﺛ','ﺜ','ﺚ'],'ج':['ﺝ','ﺞ','ﺟ','ﺠ'],'ح':['ﺡ','ﺣ','ﺤ','ﺢ'],'خ':['ﺧ','ﺨ','ﺦ'],'د':['ﺩ','ﺪ'],'ذ':['ﺬ','ﺫ'],'ر':['ﺭ','ﺮ'],'ز':['ﺯ','ﺰ',],'س':['ﺱ','ﺲ','ﺳ','ﺴ',],'ش':['ﺵ','ﺶ','ﺷ','ﺸ'],'ص':['ﺹ','ﺺ','ﺻ','ﺼ',],'ض':['ﺽ','ﺾ','ﺿ','ﻀ'],'ط':['ﻃ','ﻄ'],'ظ':['ﻅ','ﻇ','ﻈ'],'ع':['ﻉ','ﻊ','ﻋ','ﻌ',],'غ':['ﻍ','ﻏ','ﻐ',],'ف':['ﻑ','ﻒ','ﻓ','ﻔ',],'ق':['ﻕ','ﻖ','ﻗ','ﻘ',],'ل':['ﻝ','ﻞ','ﻟ','ﻠ',],'م':['ﻡ','ﻢ','ﻣ','ﻤ',],'ن':['ﻥ','ﻦ','ﻧ','ﻨ',],'چ':['ﭺ','ﭻ','ﭼ','ﭽ'],'ڈ':['ﮈ','ﮉ'],'ڑ':['ﮍ','ﮌ'],'ژ':['ﮋ',],'ک':['ﮎ','ﮏ','ﮐ','ﮑ','ﻛ','ك'],'گ':['ﮒ','ﮓ','ﮔ','ﮕ'],'ں':['ﮞ','ﮟ'],'و':['ﻮ','ﻭ','ﻮ',],'ؤ':['ﺅ'],'ھ':['ﮪ','ﮬ','ﮭ','ﻬ','ﻫ','ﮫ'],'ہ':['ﻩ','ﮦ','ﻪ','ﮧ','ﮩ','ﮨ','ه',],'ۂ':[],'ۃ':['ة'],'ء':['ﺀ'],'ی':['ﯼ','ى','ﯽ','ﻰ','ﻱ','ﻲ','ﯾ','ﯿ','ي'],'ئ':['ﺋ','ﺌ',],'ے':['ﮮ','ﮯ','ﻳ','ﻴ',],'ۓ':[],'۰':['٠'],'۱':['١'],'۲':['٢'],'۳':['٣'],'۴':['٤'],'۵':['٥'],'۶':['٦'],'۷':['٧'],'۸':['٨'],'۹':['٩'],'۔':[],'؟':[],'٫':[],'،':[],'لا':['ﻻ','ﻼ'],'':['ـ']}
    translator = {}
    for key, value in urdu_character_mapping.items():
        translator.update(dict.fromkeys(map(ord, value), key))
    return input.translate(translator)

def remove_diacritic(text: str) -> str:
    """
        This Functions returns a string after removing diacritics from it.
        
        Parameters:
                text (str): String that needs to be processed which contains words with diacritics
        Returns:
                text (string): String from which diacritics have been removed
        Examples:
                >>> text = "میرا نام اِمتیاز ہے۔"
                >>> text = remove_diacritic(text)
                >>> print(text)
                میرا نام امتیاز ہے۔
                >>> text1 = "علی کو شیرِ پنجاب بلایا جاتا ہے۔"
                >>> text1 = remove_diacritic(text1)
                >>> print(text1)
                علی کو شیر پنجاب بلایا جاتا ہے۔
    """
    diacritics = re.compile("["u"\u064e"  u"\u064B"  u"\u0670"  u"\u0650"  u"\u064F" u"\u064d""]+", re.UNICODE)
    text = diacritics.sub('', text)
    return text

def digits_space(text: str) -> str:
    """
        This Functions returns a string after putting a space in between the words and digits if they are joined in the
        actual string it.
            Parameters:
                  text (str): String that needs to be processed
            Returns:
                  text1 (string): String in which space has been included between the words and digits.
            Examples:
                >>> text = digits_space("میرا نام دانیال2089234455453 ہے۔ میں مقصد لیب 76566753میں کام کرتا ہوں")
                    'میرا نام دانیال 2089234455453 ہے۔ میں مقصد لیب 76566753 میں کام کرتا ہوں'

                >>> text1 = digits_space("وہ لاہور 20کار لے ک222 گھر گیا۔")
                    'وہ لاہور 20 کار لے ک 222 گھر گیا۔'
    """

    urdu_eng_digits = "0123456789۰۱۲۳۴۵۶۷۸۹"

    list_text = list(text)
    corrtext = ''
    for i in range(0, len(list_text)-1):

        if list_text[i] not in urdu_eng_digits and list_text[i+1] in urdu_eng_digits:
            corrtext += list_text[i]

            if list_text[i] != ' ':
                corrtext += ' '

        elif list_text[i] in urdu_eng_digits and list_text[i+1] not in urdu_eng_digits:
            corrtext += list_text[i]

            if list_text[i+1] != ' ':
                corrtext += ' '
        else:
            corrtext += list_text[i]
    corrtext += list_text[-1]
    return corrtext

def space_english_words(input: str) -> str:
    """
        This Functions returns a string after putting a space before and after english words.
            Parameters:
                  text (str): String that needs to be processed
            Returns:
                  string (str): String in which space has been included before and after english words.
            Examples:
                >>> space_english_words("میرا نام Daniyalہے۔ میں مقصد لیب میں کام کرتا ہوں")
                    'میرا نام Daniyal ہے۔ میں مقصد لیب میں کام کرتا ہوں'

                >>> text1 = digits_space("وہ لاہور twentyکار لے ک گھر گیا۔")
                    'وہ لاہور twenty کار لے ک گھر گیا۔'
    """

    all_urdu_characters = {'\u0600': '\u0600', '\u0601': '\u0601', '\u0602': '\u0602', '\u0603': '\u0603', '،': '\u060c', '؍': '\u060d', '؎': '\u060e', '؏': '\u060f', 'ؐ': '\u0610', 'ؑ': '\u0611', 'ؒ': '\u0612', 'ؓ': '\u0613', 'ؔ': '\u0614', 'ؕ': '\u0615', '؛': '\u061b', '؟': '\u061f', 'ء': '\u0621', 'ً': '\u064b', 'ٌ': '\u064c', 'ٍ': '\u064d', 'َ': '\u064e', 'ُ': '\u064f', 'ِ': '\u0650', 'ّ': '\u0651', 'ْ': '\u0652', 'ٓ': '\u0653', 'ٔ': '\u0654', 'ٖ': '\u0656', 'ٗ': '\u0657', '٘': '\u0658', '٪': '\u066a', '٫': '\u066b', '٬': '\u066c', 'ٰ': '\u0670', '۔': '\u06d4', 'آ': '\u0622', 'أ': '\u0623', 'ا': '\u0627', 'ب': '\u0628', 'پ': '\u067e', 'ت': '\u062a', 'ٹ': '\u0679', 'ث': '\u062b', 'ج': '\u062c',
                           'چ': '\u0686', 'ح': '\u062d', 'خ': '\u062e', 'د': '\u062f', 'ڈ': '\u0688', 'ذ': '\u0630', 'ر': '\u0631', 'ڑ': '\u0691', 'ز': '\u0632', 'ژ': '\u0698', 'س': '\u0633', 'ش': '\u0634', 'ص': '\u0635', 'ض': '\u0636', 'ط': '\u0637', 'ظ': '\u0638', 'ع': '\u0639', 'غ': '\u063a', 'ف': '\u0641', 'ق': '\u0642', 'ک': '\u06a9', 'گ': '\u06af', 'ل': '\u0644', 'م': '\u0645', 'ن': '\u0646', 'ں': '\u06ba', 'و': '\u0648', 'ؤ': '\u0624', 'ھ': '\u06be', 'ہ': '\u06c1', 'ۂ': '\u06c2', 'ۃ': '\u06c3', 'ی': '\u06cc', 'ئ': '\u0626', 'ے': '\u06d2', 'ۓ': '\u06d3', '۰': '\u06f0', '۱': '\u06f1', '۲': '\u06f2', '۳': '\u06f3', '۴': '\u06f4', '۵': '\u06f5', '۶': '\u06f6', '۷': '\u06f7', '۸': '\u06f8', '۹': '\u06f9', }
    space_before = re.compile(r"(?<=[" + "".join(all_urdu_characters) + "])(?=[a-zA-Z])",
                              flags=re.U | re.M | re.I)
    space_after = re.compile(r"(?<=[a-zA-Z])(?=[" + "".join(all_urdu_characters) + "])",
                             flags=re.U | re.M | re.I)
    return space_before.sub(' ', space_after.sub(' ', input))

def normalize_whitespace(input: str) -> str:
    """
        This Functions returns a string after replacing multiple white spaces with a single whitespace
            Parameters:
                  input (str): String that needs to be processed
            Returns:
                  string (str): String in which multiple whitespace has been removed from text.
            Examples:
                >>> space_english_words("وہ لاہور    لے ک گھر گیا۔")
                    'وہ لاہور لے ک گھر گیا۔'
    """

    remove_space = re.compile(r'(?!\n)[ \t]+')
    remove_new_line = re.compile(r'((\r\n)|[\n\v])+')
    return remove_space.sub(' ', remove_new_line.sub(r'\n', input)).strip()

def normalize(text: str) -> str:
    """
        This Function returns a string after normalizing it by doing the following:
        > removing extra white spaces
        > replace english digits with urdu digits
        > adding space between digits and words
        > adding space between english and urdu words
        > bringing all arabic script characters in range of urdu unicode characters
        > remove diacritics
        
        Parameters:
                text (str): String that needs to be normalized
        Returns:
                text (string): Normalized string
        Examples:
                >>> text1 = "ﺂن علی کو شیرِ پنجاب بلایا جاتا ہے ۔2022"
                >>> text1 = normalize(text1)
                >>> print(text1)
                آن علی کو شیر پنجاب بلایا جاتا ہے۔۲۰۲۲
    """
    return normalize_characters(remove_diacritic(replace_digits(space_punctuations(digits_space(space_english_words(normalize_whitespace(text)))),False)))
