import re

def replace_url(text: str, replace_with: str = "") -> str:
    """
        Returns a string after replacing url from the input text
            Parameters:
                text (str): String that needs to be processed
                replace_with (str): url which replaces the url in original text
            Returns:
                (str): String from which url(s) have been replaced
            Examples:
                >>> replace_url("بَراًےمہربانی  کچھدیر دیر http://url.com/bla1/blah1/ انتظارکیجیًے")
                'بَراًےمہربانی کچھدیر دیر انتظارکیجیًے'

                >>> replace_url("بَراًےمہربانی  کچھدیر دیر http://url.com/bla1/blah1/ انتظارکیجیًے",replace_with='https://www.google.com.sa/')
                'بَراًےمہربانی کچھدیر دیر https://www.google.com.sa/ انتظارکیجیًے'
    """

    all_urls = re.findall("(?P<url>https?://[^\s]+)", text)

    for i in all_urls:
        text = text.replace(i, replace_with)

    return text.replace('  ', ' ')

def replace_email(text: str, replace_with: str = "") -> str:
    """
        Returns a string after replacing email from the input text
            Parameters:
                text (str): String that needs to be processed
                replace_with (str): Email which replaces the email(s) in the original text
            Returns:
                (str): String from which email(s) have been replaced
            Examples:
                >>> replace_email("بَراًےمہربانی shubhamg199630@gmail.com کچھدیر دیر priya@yahoo.com انتظارکیجیًے")
                'بَراًےمہربانی کچھدیر دیر انتظارکیجیًے'

    """

    all_emails = re.findall('\S+@\S+', text)

    for i in all_emails:
        text = text.replace(i, replace_with)

    return text.replace('  ', ' ')

def replace_number(text: str, replace_with: str = "") -> str:
    """
        This Functions returns a string after replacing number(s)
            Parameters:
                  text (str): String that needs to be processed
                  replace_with (str): String containing text which replaces numbers
            Returns:
                  string (str): String in which numbers have been replaced.
            Examples:
                >>> replace_number("میں نے 10000 میں کار خریدی۔",replace_with='دس ہزار')
                    'میں نے دس ہزار میں کار خریدی۔'
    """

    all_numbers = re.findall('[0-9]+', text)

    for i in all_numbers:
        text = text.replace(i, replace_with)

    return text.replace('  ', ' ')

def replace_currency_symbols(text: str, replace_with: dict = "") -> str:
    """
        This Functions returns a string after replacing currency symbols with currency abbreviation
            Parameters:
                  text (str): String that needs to be processed
                  replace_with (dict): Dictionary containing currency symbol with abbreviation
            Returns:
                  string (str): String in which currency symbol has been replaced with abbreviation from text.
            Examples:
                >>> replace_currency_symbols("میں نے $10000 میں کار خریدی۔")
                    'میں نے USD10000 میں کار خریدی۔'
    """

    currencies = {"$": "USD", "zł": "PLN", "£": "GBP", "¥": "JPY", "฿": "THB", "₡": "CRC", "₦": "NGN",
                  "₩": "KRW", "₪": "ILS", "₫": "VND", "€": "EUR", "₱": "PHP", "₲": "PYG", "₴": "UAH", "₹": "INR"}

    if len(replace_with) == 0:
        for key, val in currencies.items():
            text = text.replace(key, val)

    else:
        for key, val in replace_with.items():
            text = text.replace(key, val)

    return text

def replace_punctuation(text: str, replace_with: str = '') -> str:
    """
      Returns a string after replacing punctuations from it.

            Parameters:
                  text (str): String that needs to be processed
                  replace_with (str): String which replaces all punctuation(s) in the text

            Returns:
                  no_punct (string): String from which punctuation have been replaced

            Examples:
                >>> replace_punctuation('میرا نام دانیال??? ہے۔ میں مقصد.... لیب میں کام کرتا ہوں۔')
                میرا نام دانیال ہے میں مقصد لیب میں کام کرتا ہوں

                >>> replace_punctuation('میرا نام علی ہے۔ میں روزانہ کام کرتا ہوں ، لیکن کیا میں کافی اچھا ہوں؟')
                میرا نام علی ہے میں روزانہ کام کرتا ہوں  لیکن کیا میں کافی اچھا ہوں

    """
    punctuations = '''!()؟-[]{}۔;:'۔"\,<>،./?@#$%^&*_~'''
    no_punct = ""

    for char in text:

        if char not in punctuations:
            no_punct += char
        else:
            no_punct += replace_with

    return no_punct.replace('  ', ' ')

def replace_english_word(text: str, replace_with: str = '') -> str:
    """
          Returns a string after removing english words from it.

                Parameters:
                      text (str): String that needs to be processed
                      replace_with (str): String which replaces all english words in the text

                Returns:
                      string (string): String from which english words have been removed


    >>> replace_english_word("ٹیسٹ نمبر one۔")
    'ٹیسٹ نمبر ۔'


    >>> replace_english_word("میرا نام Ali علی ہے۔")
    'میرا نام علی ہے۔'


    >>> replace_english_word("میں فاسٹ university me پڑہتا houn۔")
    'میں فاسٹ پڑہتا ۔'

    """

    all_english_letters = re.findall("[a-zA-Z]", text)

    mul_spa = re.compile(r'(?!\n)[ \t]+')

    for i in all_english_letters:

        text = text.replace(i, replace_with)

    return mul_spa.sub(' ', text)

def replace_emoji(text: str, replace_with: str = '') -> str:
    """
          This Functions returns a string after removing emojis from it.
          By using re.sub function which replaces the emoji with NULL character
                Arguments:
                      text (str): String that needs to be processed
                      replace_with (str): String which replaces all emojis in the text
                Returns:
                      string (string): String from which emojis have been removed
                Examples:
                      >>> replace_emoji("😡😴😢🥵🥶😠🤬😈😲😵😡😴😢🥵🥶😠🤬😈😲😵میرا نام دانیال ہے۔ میں مقصد لیب میں کام کرتا ہوں۔")
                      میرا نام دانیال ہے۔ میں مقصد لیب میں کام کرتا ہوں۔

                      >>> replace_emoji("ہم نے اس کی سالگرہ منائی🎂🏳️‍🌈۔🥳")
                      ہم نے اس کی سالگرہ منائی۔
    """

    emojis = re.compile(
        "["u"\U0001F600-\U0001F64F"  u"\U0001F300-\U0001F5FF"  u"\U0001F680-\U0001F6FF"
        u"\U0001F1E0-\U0001F1FF"  u"\U00002500-\U00002BEF"  u"\U00002702-\U000027B0"
        u"\U00002702-\U000027B0"u"\U000024C2-\U0001F251"u"\U0001f926-\U0001f937"
        u"\U00010000-\U0010ffff"u"\u2640-\u2642" u"\u2600-\u2B55"
        u"\u200d"u"\u23cf"u"\u23e9"u"\u231a"u"\ufe0f"  u"\u3030""]+",
        re.UNICODE)

    text = re.sub(emojis, replace_with, text)

    return text

def replace_mentions(string: str, replace_with: str = '') -> str:
    """
    Removes mentions from text by removing all instances of ``@...``.

    Args:
        string: text which needs to be processed
        replace_with (str): String which replaces all mentions in the text

    Returns:
        str: returns a string object containing replaced mentions.


    Examples:
        >>> replace_mentions("hello @zyena world")
        hello world
        >>> replace_mentions("@username کےسے ہو بھائ @username2 کہاں ہیں آپ")
        کےسے ہو بھائ کہاں ہیں آپ
    """
    text = ' '.join(
        re.sub("([@][A-Za-z0-9_ا-ي]+)|(\w+:\/\/\S+)", replace_with, string).split())
    return text

def replace_hashtags(text: str, replace_with: str = '') -> str:
    """
    Removes hashtags from text by replacing all instances of '#...'

    Args:
        text (str):  Text which needs to be processed
        replace_with (str): String which replaces all hashtags in the text

    Returns:
        str: returns a string object containing cleaned text.

    Examples:
        >>> replace_hashtags("weather #sunny today")
        weather today

        >>> replace_hashtags("#weather کیا خوبسورت موسم ہے آج")
        کیا خوبسورت موسم ہے آج
    """
    text = ' '.join(
        re.sub("([#][A-Za-z0-9_ا-ي]+)|(\w+:\/\/\S+)", replace_with, text).split())
    return text

def preprocess(text: str) -> str:
    """
    Removes hashtags, mentions, emojis, english words, punctuations, currency symbols, numbers, emails and url 
    from input

    Args:
        text (str):  Text which needs to be preprocessed

    Returns:
        str: returns a preprocessed string

    Examples:
        >>> preprocess("weather #sunny , @hello shubhamg199630@gmail.com 😴😢 کیا خوبسورت موسم ہے آج")
        کیا خوبسورت موسم ہے آج
    """
    return replace_url(replace_email(replace_english_word(replace_currency_symbols(
        replace_punctuation(replace_emoji(replace_mentions(replace_mentions(replace_number(text
        )))))))))
